/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlMlStaticStructLearn.hpp                                  //
//                                                                         //
//  Purpose:   The base class for all structure learning                   //
//             algorithms on complete data for static graphical models     //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#ifndef __PNLMLSTATICSTRUCTLEARN_HPP__
#define __PNLMLSTATICSTRUCTLEARN_HPP__

#include <math.h>
#include "pnlBNet.hpp"
#include "pnlDAG.hpp"
#include "pnlCPD.hpp"
#include "pnlTabularCPD.hpp"
#include "pnlGaussianCPD.hpp"
#include "pnlTreeCPD.hpp"
#include "pnlSparseMatrix.hpp"
#include "pnlStaticLearningEngine.hpp"

PNL_BEGIN

typedef PNL_API enum
{
	BIC,
	AIC,
	VAR	//not implemented	
} EScoreFunTypes;

typedef PNL_API enum
{
	StructLearnHC, //hill climbing
	StructLearnSA,  //Simulated Annealing, not implemented
	StructLearnMCMC	//Markov Chain Monte Carlo, not implemented	
} EOptimizeTypes;

class PNL_API CMlStaticStructLearn : public CStaticLearningEngine
{
public:
	
	~CMlStaticStructLearn();    
	virtual void Learn()=0;

	const CBNet* GetResultBNet()const;
	const CDAG*  GetResultDAG()const;
	const int* GetResultRenaming()const;

	float ScoreFamily(intVector vFamily);  //vFamily = [parents self]
	float ScoreFamily(int nFamily, int* vFamily, int *indexes);  //lai, 05-17-2004
	float ScoreDAG(CDAG* pDAG, floatVector* familyScore);
	float ScoreDAG(CDAG* pDAG, float* familyScore);  //lai, 06-15-2004
	void  CreateResultBNet(CDAG* pDAG);

	void  SetInitGraphicalModel(CGraphicalModel* pGrModel);
	//used to re-learn from another initial DAG. nodes order and nodes types
	//need to be the same as the original m_pGrModel. 

	EScoreFunTypes	m_ScoreType;  //lai, 05-17-2004
	
protected:
	CMlStaticStructLearn(CStaticGraphicalModel *pGrModel, ELearningTypes LearnType, 
						  EOptimizeTypes AlgorithmType, EScoreFunTypes ScoreType, int nMaxFanIn,
						  intVector& vAncestor, intVector& vDescent);

	int		m_nNodes;
	int		m_nMaxFanIn;
	intVector	m_vResultRenaming;
	intVector	m_vAncestor;
	intVector	m_vDescent;
	CBNet*		m_pResultBNet;
	CDAG*		m_pResultDAG;
	//EScoreFunTypes	m_ScoreType;
	EOptimizeTypes	m_Algorithm;

#ifdef _MPI              //lai, 06-09-2004
    int nodes_per_proc;  //lai, 06-09-2004
    int proc_startNode;  //lai, 06-09-2004
	int proc_endNode;    //lai, 06-09-2004
#endif                   //lai, 06-09-2004
	
private:
	//float ComputeFamilyScore(intVector vFamily);
	CCPD* ComputeFactor(intVector vFamily, CGraphicalModel* pGrModel, CEvidence** pEvidences);

	//CCPD* CreateRandomCPD(int nfamily, const int* family, CGraphicalModel* pGrModel);
public:
	//CSparseMatrix<float>*	m_pFamilyScoreCache;
	CSparseMatrix<float>** m_vNodeScoreCache; //lai, 05-08-2004
	float ComputeFamilyScore(intVector vFamily);  //lai, multi-threading, 05-13-2004
	float ComputeFamilyScore(int nFamily, int* vFamily);  //lai, multi-threading, 05-17-2004
	CCPD* CreateRandomCPD(int nfamily, const int* family, CGraphicalModel* pGrModel); //lai, multi-threading, 05-17-2004
};

PNL_END

#endif // __PNLMLSTATICSTRUCTLEARN_HPP__
