/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlTypeDefs.hpp                                             //
//                                                                         //
//  Purpose:   PNL's types definitions                                     //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
//  The typedef of the intVector and intVecVector types
//
//////////////////////////////////////////////////////////////////////

#ifndef __PNLTYPEDEFS_H__
#define __PNLTYPEDEFS_H__

PNL_BEGIN

#ifdef PNL_VC6_WITH_BROKEN_TEMPLATES
#define PNL_VC6
#endif

using namespace std;

enum EMargType
{
    PNL_MARG_TYPE_SUM,
    PNL_MARG_TYPE_ADD = PNL_MARG_TYPE_SUM,
    PNL_MARG_TYPE_MAX
};

enum EAccumType
{
    PNL_ACCUM_TYPE_STORE,
    PNL_ACCUM_TYPE_ACCUMULATE = -1,
    PNL_ACCUM_TYPE_SUM = PNL_MARG_TYPE_SUM,
    PNL_ACCUM_TYPE_ADD = PNL_ACCUM_TYPE_SUM,
    PNL_ACCUM_TYPE_MAX = PNL_MARG_TYPE_MAX,
    PNL_ACCUM_TYPE_MUL = 4
};

template<class T> class PNL_API GeneralAllocator: public std::allocator<T>
{
public:
    typedef typename std::allocator<T>::pointer pointer;
    typedef typename std::allocator<T>::size_type size_type;

    GeneralAllocator() {}
#ifndef PNL_VC6
    template <class TOther>
        GeneralAllocator(const std::allocator<TOther> &al) {}
#endif
    GeneralAllocator& operator=(const std::allocator<T> &al)
    {
        return *this;
    }
    bool operator==(const std::allocator<T> &al)
    {
        return false;
    }
    bool operator!=(const std::allocator<T> &al)
    {
        return true;
    }
    virtual ~GeneralAllocator() {}
    virtual pointer allocate( size_type n, const void *hint = 0 )
    {   return allocator<T>::allocate(n, hint); }
    virtual void deallocate(pointer p, size_type n)
    {   allocator<T>::deallocate(p, n); }
    virtual void construct(pointer p, const T& v)
    {   allocator<T>::construct(p, v); }
    virtual void destroy(pointer p)
    {   allocator<T>::destroy(p); }
#ifndef PNL_VC6
    template<class TOther>
        struct rebind
        {   // convert an GeneralAllocator<T> to an GeneralAllocator<TOther>
            typedef GeneralAllocator<TOther> other;
        };
#endif
};

#ifndef SWIG

#define _OPT

#ifndef _OPT
class PNL_API Value
{
public:
    int   GetInt() const { return m_tVal; }
    float GetFlt() const { return m_cVal; }
    void SetInt(int v)     { m_tVal = v; m_cVal = float(v); m_bInt = true; }
    void SetFlt(float v)   { m_tVal = int(v);   m_cVal = v; m_bInt = false; }
    void SetUnknownAsFlt(float v) { m_tVal = int(v); m_cVal = v; m_bInt = 2; }
    Value( int v ):   m_bInt(true),  m_tVal(v), m_cVal(float(v)) {}
    Value( float v ): m_bInt(false), m_tVal(int(v)), m_cVal(v) {}

    Value() {}
    char IsDiscrete() const { return m_bInt; }

private:
    char m_bInt;
    float m_cVal;
    int m_tVal;
};
#else
class PNL_API Value
{
public:
    int   GetInt() const { return u_val.m_tVal; }
    float GetFlt() const { return u_val.m_cVal; }
    void SetInt(int v)     { u_val.m_tVal = v; }
    void SetFlt(float v)   { u_val.m_cVal = v; }
    void SetUnknownAsFlt(float v) { u_val.m_cVal = v; }
    Value(int v)     { u_val.m_tVal = v; }
    Value(float v)   { u_val.m_cVal = v; }

    Value() {}
	char IsDiscrete() const { return u_val.m_bInt; }

private:
  union 
  {    //added by lai, 03-05-2004
    char m_bInt;
    float m_cVal;
    int m_tVal;
  } u_val;   //added by lai, 03-05-2004
};
#endif
#endif

#pragma warning(push, 3)
template<class Type, class Allocator = GeneralAllocator<Type> >
class PNL_API pnlVector: public std::vector<Type, Allocator>
{
public:
    typedef pnlVector<Type,Allocator> MyType;
    typedef typename std::vector<Type, Allocator> ParentType;

    typedef typename ParentType::iterator iterator;
    typedef typename ParentType::const_iterator const_iterator;
    typedef typename ParentType::allocator_type allocator_type;
    typedef typename ParentType::size_type size_type;

#if 1
    explicit pnlVector(const Allocator& Al = Allocator())
        : ParentType(Al) {}
#else
    explicit pnlVector(const Allocator& Al = Allocator())
        : ParentType(Al) {}
#endif

    explicit pnlVector( size_type N, const Type& V = Type(),
                        const Allocator& Al = Allocator() )
        : std::vector<Type, Allocator>(N, V, Al) {}
    
    pnlVector(const MyType& X)
        : std::vector<Type, Allocator>(X) {}
    
#ifndef PNL_VC6
    template< typename InputIterator >
    pnlVector( InputIterator first, InputIterator last, const Allocator &ator = Allocator() )
        : std::vector< Type, Allocator >( first, last, ator ) {}
#else
    pnlVector( const_iterator F, const_iterator L, const allocator_type &Al = Allocator() )
        : std::vector<Type, Allocator>( F, L, Al ) {}
#endif

    MyType &operator=( const MyType &X )
    {
        ParentType::operator=( static_cast< const ParentType >( X ) );
        return *this;
    }
};
#pragma warning (pop)

typedef std::queue<int>                 intQueue;

typedef pnlVector<Value>                valueVector;
typedef pnlVector<Value*>               pValueVector;
typedef pnlVector<valueVector>          valueVecVector;
typedef pnlVector<const Value*>         pConstValueVector;

typedef pnlVector<int>                  intVector;
typedef pnlVector<int *>                intPVector;
typedef pnlVector<const int*>           pConstIntVector;


typedef pnlVector< intVector>           intVecVector;

typedef pnlVector< intVector* >         intVecPVector;

typedef pnlVector<float>                floatVector;

typedef pnlVector<double>               doubleVector;

typedef pnlVector< floatVector>         floatVecVector;

typedef pnlVector<bool>                 boolVector;
typedef pnlVector< boolVector>          boolVecVector;


class CNodeType;

typedef pnlVector< CNodeType>           nodeTypeVector;

typedef pnlVector< CNodeType*>          pNodeTypeVector;

typedef pnlVector< const CNodeType*>    pConstNodeTypeVector;


class CEvidence;

typedef pnlVector< CEvidence*>          pEvidencesVector;

typedef pnlVector< const CEvidence*>    pConstEvidenceVector;

typedef pnlVector<pConstEvidenceVector> pConstEvidencesVecVector;

typedef pnlVector< pEvidencesVector>    pEvidencesVecVector;


class CFactor;

typedef pnlVector<CFactor*>             pFactorVector;

typedef pnlVector<const CFactor*>       pConstFactorVector;


class CPotential;
typedef pnlVector< const CPotential*>   pConstPotVector;

typedef pnlVector< CPotential*>         potsPVector;



typedef pnlVector< potsPVector>         potsPVecVector;


class CCPD;
typedef pnlVector<const CCPD *>         pConstCPDVector;


class CGaussianCPD;
typedef pnlVector<CGaussianCPD *>       pGaussianCPDVector;

class CDistribFun;
typedef pnlVector< CDistribFun*>        distrPVector;

PNL_END

#endif // __PNLTYPEDEFS_H__
