/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlMlStaticStructLearn.cpp                                  //
//                                                                         //
//  Purpose:   CMlStaticStructLearn class member functions implementation  //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
#include "pnlConfig.hpp"
#include "pnlMlStaticStructLearn.hpp"
#include "mpiWrap.h" 

PNL_USING
const CBNet* CMlStaticStructLearn::GetResultBNet() const
{
    return m_pResultBNet;
}

const CDAG*  CMlStaticStructLearn::GetResultDAG() const
{
    return m_pResultDAG;
}

const int* CMlStaticStructLearn::GetResultRenaming()const
{
    return &m_vResultRenaming.front();
}

CMlStaticStructLearn::CMlStaticStructLearn(CStaticGraphicalModel *pGrModel,
					   ELearningTypes LearnType, 
					   EOptimizeTypes AlgorithmType,
					   EScoreFunTypes ScoreType, int nMaxFanIn,
					   intVector& vAncestor, intVector& vDescent)
					   :CStaticLearningEngine(pGrModel, LearnType),
					   m_pResultBNet(NULL), m_pResultDAG(NULL)
{ 
    int nnodes = pGrModel->GetNumberOfNodes();
    m_nNodes = nnodes;
    int i;
    for (i = 0; i<nnodes; i++) m_vResultRenaming.push_back(i);

    m_ScoreType = ScoreType;
    m_Algorithm = AlgorithmType;
    m_nMaxFanIn = nMaxFanIn;
    m_vAncestor.assign(vAncestor.begin(), vAncestor.end());
    m_vDescent.assign(vDescent.begin(),vDescent.end());
    PNL_CHECK_RANGES(nMaxFanIn, 1, nnodes);
    int nMaxFamily = nMaxFanIn + 1;
    intVector ranges(nMaxFamily);
    for(i=0; i<nMaxFamily; i++) ranges[i] = nnodes+1;
    
	float max = (float)1e38;
 	//m_pFamilyScoreCache = CSparseMatrix<float>::Create(nMaxFamily, 
	//	                                   &ranges.front(), max, 0);
	m_vNodeScoreCache = (CSparseMatrix<float>**)malloc(m_nNodes * sizeof(CSparseMatrix<float>*));

#ifdef _MPI
    nodes_per_proc = m_nNodes / _num_procs;          //lai, 06-09-2004
    proc_startNode = nodes_per_proc * _my_rank;      //lai, 06-09-2004
    if (_my_rank == (_num_procs - 1))                //lai, 06-09-2004
     proc_endNode = m_nNodes;                        //lai, 06-09-2004
    else                                             //lai, 06-09-2004
     proc_endNode = nodes_per_proc * (_my_rank + 1); //lai, 06-09-2004
	//printf("begin allocate m_vNodeScoreCache array 0 %d %d %d\n", proc_startNode, proc_endNode, m_nNodes);
	for(i=0; i<proc_startNode; i++) m_vNodeScoreCache[i] = NULL;
	for(i=proc_startNode; i<proc_endNode; i++)
	{
		m_vNodeScoreCache[i] = CSparseMatrix<float>::Create(nMaxFanIn, 
		                                   &ranges.front(), max, 0);
	}
	for(i=proc_endNode; i<m_nNodes; i++) m_vNodeScoreCache[i] = NULL;
#else
	for(i=0; i<m_nNodes; i++)
	{
		m_vNodeScoreCache[i] = CSparseMatrix<float>::Create(nMaxFanIn, 
		                                   &ranges.front(), max, 0);
	}
#endif
	//lai, 05-08-2004
	//printf("finish allocate m_vNodeScoreCache array\n");
}

CMlStaticStructLearn::~CMlStaticStructLearn()
{
    //delete m_pFamilyScoreCache;
	for(int i=0; i<m_nNodes; i++)
		delete m_vNodeScoreCache[i];
	delete m_vNodeScoreCache;
	//lai, 05-08-2004
    if(m_pResultBNet) delete m_pResultBNet;
    if(m_pResultDAG) delete m_pResultDAG;
}

void CMlStaticStructLearn::SetInitGraphicalModel(CGraphicalModel* pGrModel)
{
    if(m_pGrModel)
        delete m_pGrModel;
    m_pGrModel = pGrModel;
}

float CMlStaticStructLearn::ScoreDAG(CDAG* pDAG, floatVector* familyScore)
{
    familyScore->clear();
    intVector vFamily;
    float fscore, score = 0.0f;
    for(int i=0; i<m_nNodes; i++)
    {
        vFamily.clear();
        pDAG->GetParents(i, &vFamily);
        vFamily.push_back(i);
        fscore = ScoreFamily(vFamily);
        familyScore->push_back(fscore);
        score += fscore;
    }
    return score;
}

float CMlStaticStructLearn::ScoreDAG(CDAG* pDAG, float* familyScore) //lai, 06-15-2004
{
    intVector vFamily;
    float fscore;
	int i;
#ifdef _MPI
	float score, pScore = 0.0f;
	float* partialfamilyScore = (float *)malloc(m_nNodes * sizeof(float));
	for(i = 0; i < m_nNodes; i ++)  partialfamilyScore[i] = 0;
    for(i=proc_startNode; i<proc_endNode; i++)
#else
	float score = 0.0f;
    for(i=0; i<m_nNodes; i++)
#endif
    {
        vFamily.clear();
        pDAG->GetParents(i, &vFamily);
        vFamily.push_back(i);
        fscore = ScoreFamily(vFamily);
#ifdef _MPI
        pScore += fscore;
		partialfamilyScore[i] = fscore;
#else
        familyScore[i] = fscore;
		score  += fscore;
#endif
    }

#ifdef _MPI
	MPI_CALL(MPI_Allreduce( &pScore, &score, 1, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD ));
	MPI_CALL(MPI_Allreduce( partialfamilyScore, familyScore, m_nNodes, MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD ));
	delete(partialfamilyScore);
#endif

    return score;
}


//float CMlStaticStructLearn::ScoreFamily(intVector vFamily,int *hitormiss)
float CMlStaticStructLearn::ScoreFamily(intVector vFamily)
{
    //int nFamily = vFamily.size();
    //int nMaxFamily = m_nMaxFanIn + 1;
    //PNL_CHECK_RANGES(nFamily, 1, nMaxFamily);

    int nParents = vFamily.size() - 1;
    PNL_CHECK_RANGES(nParents, 0, m_nMaxFanIn);
	int *indexes=new int[m_nMaxFanIn];

    //intVector indexes(nMaxFamily);
    int i;
    //for(i=0; i<nMaxFamily; i++) indexes[i] = 0;
    //for(i=0; i<nFamily; i++)
    //{
    //    indexes[i] = vFamily[i]+1;
    //}
    for(i=0; i<nParents; i++)
    {
        indexes[i] = vFamily[i]+1;
    }
    for(i=nParents; i<m_nMaxFanIn; i++)
    {
        indexes[i] = 0;
    }
	int node = vFamily[nParents];

	float score;
	float* existedscore;
//#ifdef _OPENMP
//#pragma omp critical (m_pFamilyScoreCache)
//#endif
	existedscore = m_vNodeScoreCache[node]->PointerOfElement(indexes, &score);
	if (!existedscore)
    {
        score = ComputeFamilyScore(vFamily);
//#ifdef _OPENMP
//#pragma omp critical (m_pFamilyScoreCache)
//#endif
		m_vNodeScoreCache[node]->SetElementByIndexes(score, indexes);//added by lai, 05-08-2004
    }
	delete []indexes;  //added by lai, 05-08-2004
	//VTPause();
    return score;
}


//lai, 05-17-2004, new ScoreFamily Functions
float CMlStaticStructLearn::ScoreFamily(int nFamily, int* vFamily, int *indexes)
{
    //VTPause();
	//VTResume();
	int nParents = nFamily - 1;
    //PNL_CHECK_RANGES(nParents, 0, m_nMaxFanIn);
    int i;
    for(i=0; i<nParents; i++)
    {
        indexes[i] = vFamily[i]+1;
    }
    for(i=nParents; i<m_nMaxFanIn; i++)
    {
        indexes[i] = 0;
    }
	int node = vFamily[nParents];

	//VTResume();
	//VTPause();
    float score;
	float* existedscore;
	existedscore = m_vNodeScoreCache[node]->PointerOfElement(indexes, &score);
	if (!existedscore)
    {
        score = ComputeFamilyScore(nFamily, vFamily);
		m_vNodeScoreCache[node]->SetElementByIndexes(score, indexes);//added by lai, 05-08-2004
    }
    return score;
}
//lai, 05-17-2004, new ScoreFamily Functions

float CMlStaticStructLearn::ComputeFamilyScore(intVector vFamily)
{
	//VTPause();
	//VTResume();
    int nFamily = vFamily.size();
	CCPD* iCPD = this->CreateRandomCPD(nFamily, &vFamily.front(), m_pGrModel);
    int ncases = m_Vector_pEvidences.size();
    iCPD->UpdateStatisticsML( &m_Vector_pEvidences.front(), ncases );
    float score = iCPD->ProcessingStatisticalData(ncases);
    int dim = iCPD->GetNumberOfFreeParameters();
    switch (m_ScoreType)
    {
    case BIC :
		score -= 0.5f * float(dim) * float(log(float(ncases)));
		break;
    case AIC :
		score -= 0.5f * float(dim);
		break;
    case VAR :
		PNL_THROW(CNotImplemented, 
			"This type score function has not been implemented yet");
		break;
    default:
		PNL_THROW(CNotImplemented, 
			"This type score function has not been implemented yet");
		break;
    }
    delete iCPD;
    return score;
}

//lai, new ComputeFamilyScore function, 05-17-2004
float CMlStaticStructLearn::ComputeFamilyScore(int nFamily, int* vFamily)
{
	CCPD* iCPD = this->CreateRandomCPD(nFamily, vFamily, m_pGrModel);
    int ncases = m_Vector_pEvidences.size();
    iCPD->UpdateStatisticsML( &m_Vector_pEvidences.front(), ncases );
    float score = iCPD->ProcessingStatisticalData(ncases);
    int dim = iCPD->GetNumberOfFreeParameters();
    switch (m_ScoreType)
    {
    case BIC :
		score -= 0.5f * float(dim) * float(log(float(ncases)));
		break;
    case AIC :
		score -= 0.5f * float(dim);
		break;
    case VAR :
		PNL_THROW(CNotImplemented, 
			"This type score function has not been implemented yet");
		break;
    default:
		PNL_THROW(CNotImplemented, 
			"This type score function has not been implemented yet");
		break;
    }
    delete iCPD;
    return score;
}
//lai, new ComputeFamilyScore function, 05-17-2004

void  CMlStaticStructLearn::CreateResultBNet(CDAG* pDAG)
{
    int i, j, k, ns;
    int nnodes = m_nNodes;
    CDAG* iDAG = pDAG->TopologicalCreateDAG(m_vResultRenaming);
    nodeTypeVector vnt;
    m_pGrModel->GetNodeTypes(&vnt);
    intVector na(nnodes);
    const int* nas = m_pGrModel->GetNodeAssociations();
    for(i=0; i<nnodes; i++) na[i] = nas[m_vResultRenaming[i]];
    m_pResultBNet = CBNet::Create(nnodes, vnt.size(), &vnt.front(), 
		                    &na.front(), static_cast<CGraph*>(iDAG));
    const CNodeType* nt;

    int nEv = m_Vector_pEvidences.size();
    CEvidence** pEv = new CEvidence*[nEv];
    intVector obsnodes(nnodes);
    for(i=0; i<nnodes; i++) obsnodes[i] = i;
    valueVector new_data;
    const Value* val;
    for(i = 0 ; i < nEv; i++)
    {
	for(j=0; j<nnodes; j++)
	{
	    val = m_Vector_pEvidences[i]->GetValue(m_vResultRenaming[j]);
		nt = m_pResultBNet->GetNodeType(j);
	    if(nt->IsDiscrete())
	    {
		    new_data.push_back(*val);
	    }
	    else
	    {
		    ns = nt->GetNodeSize();
		    for(k=0; k<ns; k++)
		        new_data.push_back(*(val+k));
	    }
	}			
	pEv[i] = CEvidence::Create(m_pResultBNet, nnodes, &obsnodes.front(), new_data);
	new_data.clear();
    }
    vnt.clear();
    
    intVector vFamily;
    m_pResultBNet->AllocFactors();
    for(i=0; i<nnodes; i++)
    {
	    vFamily.clear();
	    iDAG->GetParents(i, &vFamily);
	    vFamily.push_back(i);
	    CCPD* iCPD = ComputeFactor(vFamily, m_pResultBNet, pEv);
	    m_pResultBNet->AttachFactor(iCPD);
    }
    
    for(i=0; i<nEv; i++)delete pEv[i];
    delete[] pEv;
}

CCPD*
CMlStaticStructLearn::ComputeFactor(intVector vFamily, CGraphicalModel* pGrModel, CEvidence** pEvidences)
{
    int nFamily = vFamily.size();
	CCPD* iCPD = this->CreateRandomCPD(nFamily, 
		             &vFamily.front(), pGrModel);
    int ncases = m_Vector_pEvidences.size();
    iCPD->UpdateStatisticsML( pEvidences, ncases );
    iCPD->ProcessingStatisticalData(ncases);
    return iCPD;
}

CCPD* 
CMlStaticStructLearn::CreateRandomCPD(int nfamily, const int* family, CGraphicalModel* pGrModel)
{
	int child = family[nfamily-1];
	CModelDomain* pMD = pGrModel->GetModelDomain();
	CFactor* factor = pGrModel->GetFactor(child);
	EDistributionType dt = factor->GetDistributionType();
	CCPD* pCPD;

	switch (dt)
    {
    case dtTabular :
		pCPD = CTabularCPD::Create(family, nfamily, pMD);
		pCPD->CreateAllNecessaryMatrices(1);
		break;
    
	case dtTree :
		pCPD = CTreeCPD::Create(family, nfamily, pMD);
		break;

    case dtGaussian :
	case dtCondGaussian :
		pCPD = CGaussianCPD::Create(family, nfamily, pMD);
		pCPD->CreateAllNecessaryMatrices(1);
		break;
    default:
		PNL_THROW(CNotImplemented, "This type of distribution has not been implemented yet");
		break;
    }
	return pCPD;
}
