#include "brLenOptEM.h"
#include "computeUpAlg.h"
#include "computeDownAlg.h"
#include "computePijComponent.h"
#include "logFile.h"
#include "likelihoodComputation.h"
#include <iostream>
#include "countTableComponent.h"
#include "computeCounts.h"
#include "fromCountTableComponentToDistance.h"
#include "../mpiWrap.h"
#include <cmath>
#include "../wtime.h"

using namespace treeInterface;
using namespace likelihoodComputation;

extern MDOUBLE BBL_para_time;

//#define VERBOS

/*
void fillCountTableComponentGam(countTableComponentGam& ctcGam,
								const stochasticProcess& sp,
								const sequenceContainer1G& sc,
								computePijGam& pij0,
								const suffStatGlobalGam& cup,
								const suffStatGlobalGam& cdown,
								const Vdouble * weights,
								tree::nodeP nodeSon,
								const Vdouble& posProbVec) {
	ctcGam.countTableComponentAllocatePlace(sp.alphabetSize(),sp.categories());
	for (int rateCat =0; rateCat< sp.categories(); ++ rateCat) {
		fillCountTableComponentGamSpecRateCategor(rateCat,ctcGam[rateCat],sp,
													sc,pij0[rateCat],
													cup,cdown,weights,posProbVec,nodeSon);
	}
}*/
void revaluateBranchLength(computePijGam& pij0,
						   const sequenceContainer1G& sc,
						const stochasticProcess& sp,
						const Vdouble * weights,
						const tree& et,
						const suffStatGlobalGam& cup,
						const suffStatGlobalGam& cdown,
						const Vdouble& posProbVec,
						MDOUBLE tollForPairwiseDist
						)
{// = 0.001
	vector<countTableComponentGam> vec; // for each node.
	vec.resize(et.iNodes());
	treeIterDownTopConst tIt(et);
	computeCounts cc; // just container of functions;
	tree::nodeP mynode = tIt.first();

#ifdef _MPI
	int nnode = 0;
	for (; mynode != tIt.end(); mynode = tIt.next()) {
		if (!tIt->isRoot()) { 
			cc.fillCountTableComponentGam(vec[mynode->id()],sp,sc,pij0,cup,cdown,weights,mynode,posProbVec);
			++nnode;
		}
	}
	int cate, alpha1, alpha2;
	int csize = sp.categories();
	int asize = pij0.alphabetSize();
	int factor2 = asize*asize;
	int factor1 = csize*factor2;
	int total = nnode * factor1;
	int num = 0;
	MDOUBLE *tmp = new MDOUBLE[total];
	MDOUBLE *sum = new MDOUBLE[total];
	for (mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
		if (!tIt->isRoot()) {
			for (cate = 0; cate < csize; ++cate)
				for (alpha1 = 0; alpha1 < asize; ++alpha1)
					for (alpha2 = 0; alpha2 < asize; ++alpha2)
						tmp[num*factor1+cate*factor2+alpha1*asize+alpha2] = vec[mynode->id()].getCounts(alpha1, alpha2, cate);
			++num;
		}
	MPI_Allreduce(tmp, sum, total, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	num = 0;
	for (mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
		if (!tIt->isRoot()) {
			for (cate = 0; cate < csize; ++cate)
				for (alpha1 = 0; alpha1 < asize; ++alpha1)
					for (alpha2 = 0; alpha2 < asize; ++alpha2)
						vec[mynode->id()].setCount(alpha1, alpha2, cate,sum[num*factor1+cate*factor2+alpha1*asize+alpha2]);
			++num;
		}
	delete[] tmp;
	delete[] sum;
	int nmean, nleft, startp;
	nmean = (int)(nnode / _num_procs);
	nleft = nnode % _num_procs;
	if (_my_rank < nleft) 
		startp = _my_rank * (++nmean);
	else
		startp = _my_rank * nmean + nleft;
	num = 0;
	tmp = new MDOUBLE[nnode];
	sum = new MDOUBLE[nnode];
	for (mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) 
		if (!tIt->isRoot()) {
			if (num < startp) {tmp[num++] = 0; continue;}
			if (num >= startp + nmean) {tmp[num++] = 0; continue;}
			fromCountTableComponentToDistance from1(vec[mynode->id()],sp,tollForPairwiseDist,mynode->dis2father());
			from1.computeDistance();
			tmp[num++] = from1.getDistance();
		}
	MPI_Allreduce(tmp, sum, nnode, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
	num = 0;
	for (mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next())
		if (!tIt->isRoot()) {
			mynode->setDisToFather(sum[num++]);
		}	
	delete[] tmp;
	delete[] sum;
#else
#ifdef _OPENMP
#pragma intel omp parallel taskq shared(mynode)
#endif
	{
		for (; mynode != tIt.end(); mynode = tIt.next()) {
		  if (!tIt->isRoot()) {
#ifdef _OPENMP		  	
#pragma intel omp task captureprivate(mynode)
#endif
		 	{
			cc.fillCountTableComponentGam(vec[mynode->id()],sp,sc,pij0,cup,cdown,weights,mynode,posProbVec);
			fromCountTableComponentToDistance from1(vec[mynode->id()],sp,tollForPairwiseDist,mynode->dis2father());
			from1.computeDistance();
			mynode->setDisToFather(from1.getDistance());
		  	}
		  }
		}
  	}
#endif
}

MDOUBLE brLenOptEM::optimizeBranchLength1G_EM(tree& et,
									const sequenceContainer1G& sc,
									const stochasticProcess& sp,
									const Vdouble * weights,
									const int maxIterations,
									const MDOUBLE epsilon,
									const MDOUBLE tollForPairwiseDist) {
	// MN - debug the segfoult down here
	
	LOG(150,<<"optimizeBranchLength1G_EM"<<endl);
	LOGDO(150,et.output(myLog::LogFile(),tree::ANCESTORID));
	LOG(150,<<"\n\n"<<endl);
	if (et.WithBranchLength() == false) et.createFlatLengthMatrix();
	MDOUBLE newL=VERYSMALL;
	int it;
	Vdouble posProb;
	suffStatGlobalGam cup;
	suffStatGlobalGam cdown;
#ifdef _MPI
	cup.allocatePlace(mpi_mean, sp.categories(), et.iNodes(), sc.alphabetSize());
	cdown.allocatePlace(mpi_mean, sp.categories(), et.iNodes(), sc.alphabetSize());
#else
	cup.allocatePlace(sc.seqLen(),sp.categories(),et.iNodes(),sc.alphabetSize());
	cdown.allocatePlace(sc.seqLen(),sp.categories(),et.iNodes(),sc.alphabetSize());
#endif
	computePijGam pij0;
//	computePijGam pij1;
//	computePijGam pij2;

	pij0.fillPij(et,sp,sc.alphabetSize(),0);
//	pij1.fillPij(et,sp,sc.alphabetSize(),1);
//	pij2.fillPij(et,sp,sc.alphabetSize(),2);

	computeUpAlg cupAlg;
	computeDownAlg cdownAlg;

#ifdef _OPT
	seqContainerTreeMap sctm(sc,et);
#endif

	int pos;
	int categor;

	MDOUBLE para_time1 = 0.0;
	timer_clear(21);
	timer_start(21);

#ifdef _MPI
	for (pos = 0; pos < mpi_mean; ++pos) {
		for (categor = 0; categor < sp.categories(); ++categor) {
#ifndef _OPT
			cupAlg.fillComputeUp(et,sc,pos + mpi_spos,pij0[categor],cup[pos][categor]);
#else
			cupAlg.fillComputeUp(et,sc,pos + mpi_spos,pij0[categor],cup[pos][categor],sctm);
#endif
			cdownAlg.fillComputeDown(et,sc,pos + mpi_spos,pij0[categor],cdown[pos][categor],cup[pos][categor]);
		}
	}
#else

#ifdef _OPENMP
	int dx1 = sc.seqLen();
	int dx2 = sp.categories();
#pragma omp parallel for private(categor)
	for (pos = 0; pos < dx1; ++pos) {
		for (categor = 0; categor < dx2; ++categor) {
#else
	for (pos = 0; pos < sc.seqLen(); ++pos) {
		for (categor = 0; categor < sp.categories(); ++categor) {
#endif

#ifndef _OPT
			cupAlg.fillComputeUp(et,sc,pos,pij0[categor],cup[pos][categor]);
#else
			cupAlg.fillComputeUp(et,sc,pos,pij0[categor],cup[pos][categor],sctm);
#endif
			cdownAlg.fillComputeDown(et,sc,pos,pij0[categor],cdown[pos][categor],cup[pos][categor]);
		}
	}
#endif

#ifdef _MPI
	MDOUBLE oldL;
	MDOUBLE my_oldL = likelihoodComputation::getTreeLikelihoodFromUp2(et,sc,sp,cup,posProb,weights);
	MPI_Allreduce(&my_oldL, &oldL, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
	MDOUBLE oldL = likelihoodComputation::getTreeLikelihoodFromUp2(et,sc,sp,cup,posProb,weights);
#endif

	timer_stop(21);
	para_time1 += timer_read(21);

	if (maxIterations==0) return oldL;
	LOG(50,<<"(bbl) startingL= "<<oldL<<endl);
#ifdef VERBOS
	cerr<<"(bbl) startingL= "<<oldL<<endl;
#endif
	for (it = 0; it < maxIterations; ++it) {
//		it =1; cerr<<"starting with the starting tree without it =0 !!!";

//		LOG(50,<<"(bbl)x it= "<<it<<" newL = "<<newL<<endl);
//
//		LOG(1,<<"tmp tree, bbl it:"<<it<<endl);
//		LOGDO(1,_et.output(myLog::LogFile(),tree::PHYLIP));
		
		tree oldT = et;
		
		timer_clear(21);
		timer_start(21);
		
		revaluateBranchLength(pij0,sc,sp,weights,et,cup,cdown,posProb,tollForPairwiseDist);

		timer_stop(21);
		para_time1 += timer_read(21);

		pij0.fillPij(et,sp,sc.alphabetSize(),0);
//		pij1.fillPij(et,sp,sc.alphabetSize(),1);
//		pij2.fillPij(et,sp,sc.alphabetSize(),2);

#ifdef _OPT
		seqContainerTreeMap sctm(sc,et);
#endif

		timer_clear(21);
		timer_start(21);

#ifdef _MPI
		for (pos = 0; pos < mpi_mean; ++pos) {
			for (categor = 0; categor < sp.categories(); ++categor) {
#ifndef _OPT
			cupAlg.fillComputeUp(et,sc,pos + mpi_spos,pij0[categor],cup[pos][categor]);
#else
			cupAlg.fillComputeUp(et,sc,pos + mpi_spos,pij0[categor],cup[pos][categor],sctm);
#endif
			cdownAlg.fillComputeDown(et,sc,pos + mpi_spos,pij0[categor],cdown[pos][categor],cup[pos][categor]);
			}
		}
#else

#ifdef _OPENMP
		int dx1 = sc.seqLen();
		int dx2 = sp.categories();
#pragma omp parallel for private(categor)
		for (int pos = 0; pos < dx1; ++pos) {
			for (int categor = 0; categor < dx2; ++categor) {
#else
		for (int pos = 0; pos < sc.seqLen(); ++pos) {
			for (int categor = 0; categor < sp.categories(); ++categor) {
#endif		

#ifndef _OPT
				cupAlg.fillComputeUp(et,sc,pos,pij0[categor],cup[pos][categor]);
#else
				cupAlg.fillComputeUp(et,sc,pos,pij0[categor],cup[pos][categor],sctm);
#endif
				cdownAlg.fillComputeDown(et,sc,pos,pij0[categor],cdown[pos][categor],cup[pos][categor]);
			}
		}
#endif

#ifdef _MPI
		MDOUBLE newL;
		MDOUBLE my_newL = likelihoodComputation::getTreeLikelihoodFromUp2(et,sc,sp,cup,posProb,weights);
		MPI_Allreduce(&my_newL, &newL, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
		MDOUBLE newL = likelihoodComputation::getTreeLikelihoodFromUp2(et,sc,sp,cup,posProb,weights);
#endif
		timer_stop(21);
		para_time1 += timer_read(21);	

#ifdef VERBOS
		cerr<<"(bbl) newL= "<<newL<<endl;
#endif

#ifndef _EVAL
		if (((newL < oldL + epsilon) && (it>0)) || (maxIterations == it)){
#else
		if(it == 19){    //ouyang: BBL run 20 iterations
#endif

			BBL_para_time += para_time1;
#ifdef _OUTPUT
			printf("BBL iterations: %d\n",it+1); 
			printf ("\nBBL para time: %.2fs\n", para_time1);
#endif
			if (newL < oldL) {
				et = oldT;
				return oldL;
			}
			else {
				return newL;
			}
		} else {oldL = newL;}
		LOG(50,<<"(bbl)y it= "<<it<<" newL = "<<newL<<endl);
#ifdef VERBOS
		cerr<<"(bbl)y it= "<<it<<" newL = "<<newL<<endl;
#endif
	}
	return newL;

}
