#include <cassert>
#include "split.h"

split::split (const int max): _max(max), _reverse(true){
    for(int j=0;j<max;++j)
      _set[1].insert(j);
}


bool split::isMember(const int key)
  {
    return(_set[_reverse].find(key)!=_set[_reverse].end());
  }

void split::add(const int newnum){	// add new member
  assert(newnum<_max && newnum >= 0);
  _set[0].insert(newnum);
  _set[1].erase(newnum);
  if (newnum==0)		// if we add "1", we may need to reverse the split
    if (_reverse){
      reverse();
    }
}

void split::remove(const int oldNum){	// add new member
    _set[0].erase(oldNum);
    _set[1].insert(oldNum);
    if (oldNum==0)		//  again, a "1" will reverse the split
      if (!_reverse){
	reverse();
      }
  }

int split::size() const  {return min(_set[_reverse].size(),_max-_set[_reverse].size());}

void split::print(ostream& sout) const{
    sout <<"["<<size()<<":"<<_reverse<<"]   ";
/////////////////////////////edited by dengxi August 9------------00
    for (set<int>::const_iterator i=*((set<int>::const_iterator *)&(_set[_reverse].begin()));i != _set[_reverse].end();++i)
      sout << *i << " ";
    sout <<" | ";
    for (set<int>::const_iterator i=*((set<int>::const_iterator *)&(_set[!_reverse].begin()));i != _set[!_reverse].end();++i)
////////////////////////////end edit------------------------------00
      sout << *i << " ";
    //    sout << endl;
  }

bool split::lessThen(const split& other) const{
    return(_set[_reverse]<other._set[other._reverse]);
  }

bool split::compatible(const split& other) const {
    {
/////////////////////////////edited by dengxi August 9------------01
    set<int>::iterator i     (*((set<int>::iterator *)&(_set[_reverse].begin())));
    set<int>::iterator i_end (*((set<int>::iterator *)&(_set[_reverse].end())));
    set<int>::iterator j     (*((set<int>::iterator *)&(other._set[other._reverse].begin())));
    set<int>::iterator j_end (*((set<int>::iterator *)&(other._set[other._reverse].end())));
////////////////////////////end edit------------------------------01

    bool igtj=false;
    bool jgti=false;
    for (;(i!=i_end && j!=j_end) && !(igtj && jgti);){
      if (*i==*j) {++i;++j;}
      else if (*i>*j) {
	jgti=true;
	++j;
      }
      else if (*i<*j) {
	igtj=true;
	++i;
      }
    }
    if (i==i_end && j!=j_end) {
      jgti=true;
    }
    else 
      if (j==j_end && i!=i_end) {
      igtj=true;
      }
    if (!(igtj && jgti))
      return true;	// see that we do not have two incompatibel sets
    //    else
    }
    {
/////////////////////////////edited by dengxi August 9------------02
      set<int>::iterator i     (*((set<int>::iterator *)&(_set[!_reverse].begin())));
      set<int>::iterator i_end (*((set<int>::iterator *)&(_set[!_reverse].end())));
      set<int>::iterator j     (*((set<int>::iterator *)&(other._set[!other._reverse].begin())));
      set<int>::iterator j_end (*((set<int>::iterator *)&(other._set[!other._reverse].end())));
////////////////////////////end edit------------------------------02

      for (;i!=i_end && j!=j_end;){
	if (*i==*j) {return (false);}
	else {
	  if (*i>*j) ++j;
	  else       ++i;
	}
      }
    }
    return (true);
  }

void split::reverse(){		// actualy reverse membership in the set
    _reverse=!_reverse;
  }

bool operator<(const split& a, 
	       const split& b) {
  return(a.lessThen(b));
}

ostream& operator<< (ostream &sout,  const split& split) {split.print(sout); return sout;}

