#include "semphySearchBestTree.h"
#include "wtime.h"

//HERE IS THE MAIN CODE FOR THE SEMPHY SEARCH
//THINGS TO TAKE INTO ACCOUNT:
//(1) Several starting point
//(2) Tmp handling for semphy-ratchet and semphy-anneal.

extern MDOUBLE BBL_para_time;
extern MDOUBLE Semphy_para_time;
extern MDOUBLE BBL_time;
extern MDOUBLE Semphy_time;

semphySearchBestTree::semphySearchBestTree(sequenceContainer1G& sc,
								  tree& startTree,
								  stochasticProcess& sp,
								  ostream& out,
								  const int numOfRandomStart) {
	if (numOfRandomStart == 1) {
		semphyBasicSearchBestTree(sc,startTree,sp,out);
	} else {
		semphyBasicSearchBestTreeManyRandomStarts(sc,startTree,sp,out,numOfRandomStart);
	}

}

#include "uniDistribution.h"
#include "aaJC.h"
#include "nucJC.h"
#include "pijAccelerator.h"
#include "trivialAccelerator.h"
#include "chebyshevAccelerator.h"
#include "stochasticProcess.h"
#include "nucleotide.h"
#include "amino.h"
#include "sequenceContainer1G.h"
#include "phylipFormat.h"
#include "maseFormat.h"
#include "fastaFormat.h"
#include "treeInterface.h"
#include "likelihoodComputation.h"
#include "computePijComponent.h"
#include "seqContainerTreeMap.h"
#include "gammaDistribution.h"
#include "sequenceContainerNG.h"
#include "brLenOptEM.h"
#include "fromInstructionFile.h"
#include "NNiProp.h"
#include "NNiSep.h"
#include "jcDistance.h"
#include "distanceTable.h"
#include "nj.h"
#include "brLenOptEM.h"
#include "semphyStep.h"
#include "mpiWrap.h"

#include "computeUpAlg.h"
#include "computeDownAlg.h"
#include "computeMarginalAlg.h"
using namespace std;
using namespace treeInterface;
using namespace likelihoodComputation;
using namespace brLenOptEM;

#include "getRandomWeights.h"
#include "wtime.h"
#include <time.h>

MDOUBLE semphySearchBestTree::semphyBasicSearchBestTreeManyRandomStarts(
								  sequenceContainer1G& sc,
								  tree& et,
								  stochasticProcess& sp,
								  ostream& out,
								  const int nRanStart) {
	MDOUBLE bestL = semphyBasicSearchBestTree(sc,et,sp);
	tree bestT = et;

	const MDOUBLE tmpForStartingTreeSearch = 0.1; // PUT IN THE OPTION FILE IF PROVE TO BE USEFULL...
	for (int it=0; it < nRanStart; ++it) {
	  LOG(3,<<"\n\n\n================================================================================================================\n\n\n"<<endl);
	  LOG(3,<<"best score so far: "<<bestL<<endl);
	  LOG(3,<<"starting random start number: "<<it+1<<endl);
		Vdouble startingTreeWeights(sc.seqLen(),0.0);
 		getRandomWeights::randomWeightsGamma(startingTreeWeights,tmpForStartingTreeSearch);
	
	
		jcDistance pd1(sc.alphabetSize()); // assumes all alphabets are the same size
		VVdouble disTab;
		vector<string> vNames;
		giveDistanceTable(&pd1,
						   sc,
						   disTab,
						   vNames,
						   &startingTreeWeights);
		NJalg nj1;
		et = nj1.computeNJtree(disTab,vNames);
		LOG(3,<<"starting random tree for this semphy iteration "<<endl);
		LOGDO(3,et.output(myLog::LogFile()));
		MDOUBLE newL = semphyBasicSearchBestTree(sc,et,sp);
		LOG(3,<<" with this random tree the log likelihood is: "<<newL<<endl);

		if (newL > bestL) {
		  LOG(3,<<" new best tree: score = "<<newL<<endl);

			bestT = et;
			bestL = newL;
			LOGDO(3,bestT.output(myLog::LogFile()));
		}
	}
	
	out<<"Semphy best tree: score = "<<bestL<<endl;
	bestT.output(out);
	return bestL;
}

MDOUBLE semphySearchBestTree::semphyBasicSearchBestTree(
								  sequenceContainer1G& sc,
								  tree& et,
								  stochasticProcess& sp) {
  
  return semphyBasicSearchBestTree(sc,et,sp,cout);
}



MDOUBLE semphySearchBestTree::semphyBasicSearchBestTree(
								  sequenceContainer1G& sc,
								  tree& et,
								  stochasticProcess& sp,
								  ostream& out) {
	// this is the most basic semphy search 
	//  - no anneal
	//	- no multiple random starts
	// start by optimizing the br-len of the input tree.
//	optimizeBranchLength1G_EM(et,sc,sp,NULL);
	double  duration;

	MDOUBLE para_time2 = 0.0;
	BBL_time = 0.0;
	Semphy_time = 0.0;
	BBL_para_time = 0.0;
	Semphy_para_time = 0.0;
	
	timer_start(1);	// Semphy step;

#ifndef _EVAL	
	for (int z=0; z < 10; ++z) {//1000
#else
	for (int z=0; z < 1; ++z) {
#endif
		tree bestTree = et;
#ifdef _OUTPUT
		LOG(3,<<"semphy iteration: "<<z<<endl);
		out<<"semphy iteration: "<<z<<endl;  //ouyang
#endif
		computePijGam pij0;
		pij0.fillPij(et,sp,sc.alphabetSize());

		suffStatGlobalGam cup;
		suffStatGlobalGam cdown;
		suffStatGlobalGam cmarg;

#ifdef _MPI
		cup.allocatePlace(mpi_mean, sp.categories(),et.iNodes(),sc.alphabetSize());
		cdown.allocatePlace(mpi_mean,sp.categories(),et.iNodes(),sc.alphabetSize());
		cmarg.allocatePlace(mpi_mean,sp.categories(),et.iNodes(),sc.alphabetSize());
#else
		cup.allocatePlace(sc.seqLen(),sp.categories(),et.iNodes(),sc.alphabetSize());
		cdown.allocatePlace(sc.seqLen(),sp.categories(),et.iNodes(),sc.alphabetSize());
		cmarg.allocatePlace(sc.seqLen(),sp.categories(),et.iNodes(),sc.alphabetSize());
#endif

		computeUpAlg cupAlg;
		computeDownAlg cdownAlg;
		computeMarginalAlg cmargAlg;
		int pos = 0;
		int categor = 0;

#ifdef _OPT
		seqContainerTreeMap sctm(sc,et);
#endif

		timer_start(2);	// fill compute up/down and so on

#ifdef _MPI
		for (pos = 0; pos < mpi_mean; ++pos) {
			for (categor = 0; categor < sp.categories(); ++categor) {
#ifdef _OPT
				cupAlg.fillComputeUp(et,sc,pos+mpi_spos,pij0[categor],cup[pos][categor],sctm);
#else
				cupAlg.fillComputeUp(et,sc,pos+mpi_spos,pij0[categor],cup[pos][categor]);
#endif				
				cdownAlg.fillComputeDown(et,sc,pos+mpi_spos,pij0[categor],cdown[pos][categor],cup[pos][categor]);
#else									
#ifdef _OPENMP		
		int tmpdx, tmpdx1;
		tmpdx = sc.seqLen();
		tmpdx1 = sp.categories();
#pragma omp parallel for private(categor)
		for (pos = 0; pos < tmpdx; ++pos) {
			for (categor = 0; categor < tmpdx1; ++categor) {
#else
		for (pos = 0; pos < sc.seqLen(); ++pos) {
			for (categor = 0; categor < sp.categories(); ++categor) {
#endif

#ifdef _OPT
				cupAlg.fillComputeUp(et,sc,pos,pij0[categor],cup[pos][categor],sctm);
#else
				cupAlg.fillComputeUp(et,sc,pos,pij0[categor],cup[pos][categor]);
#endif				
				cdownAlg.fillComputeDown(et,sc,pos,pij0[categor],cdown[pos][categor],cup[pos][categor]);
#endif
			}
		}

		LOG(3,<<"starting semphy step by bbl...");
		

		Vdouble posLike;
#ifdef _MPI
		MDOUBLE tLike;		
		MDOUBLE my_tLike = likelihoodComputation::getTreeLikelihoodFromUp2(et,
							sc,
							sp,
							cup,
							posLike, // fill this vector with each position likelihood but without the weights.
							NULL);//const Vdouble * weights=0);
		MPI_Allreduce(&my_tLike, &tLike, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
		MDOUBLE tLike = likelihoodComputation::getTreeLikelihoodFromUp2(et,
							sc,
							sp,
							cup,
							posLike, // fill this vector with each position likelihood but without the weights.
							NULL);//const Vdouble * weights=0);
#endif		
		
		VVdouble ratePosteriorProb;
#ifdef _MPI
		MDOUBLE tLike2;
		MDOUBLE my_tLike2 = likelihoodComputation::getPosteriorOfRates(et,
							sc,
							sp,
							cup,
							ratePosteriorProb, // fill this vector with each position likelihood but without the weights.
							NULL);//const Vdouble * weights=0);
		MPI_Allreduce(&my_tLike2, &tLike2, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
		MDOUBLE tLike2 = likelihoodComputation::getPosteriorOfRates(et,
							sc,
							sp,
							cup,
							ratePosteriorProb, // fill this vector with each position likelihood but without the weights.
							NULL);//const Vdouble * weights=0);
#endif

		// commented out by matan, 4-2-03
// 		{ // very checking
// 			for (int k=0; k < ratePosteriorProb[4].size(); ++k ) {
// 				cerr<<"ratePosteriorProb["<<k<<"]= "<<ratePosteriorProb[4][k]<<endl;
// 			}
//		}

#ifdef _OPENMP
		MDOUBLE ompdelta = 0.000001;
		if (abs(tLike - tLike2) > ompdelta) {
#else
		if (tLike != tLike2) {
#endif		
			cerr<<" tLike = "<<tLike<<endl;
			cerr<<" tLike2 = "<<tLike2<<endl;
			errorMsg::reportError(" error in computational - some of posterior != prob... ");
		}

#ifdef _OUTPUT	
		LOG(3,<<"starting likelihood = "<<tLike<<endl);
		out<<"starting likelihood= "<<tLike<<endl;  //ouyang
#endif
		MDOUBLE oldL = tLike;

#ifdef _MPI
		for (pos = 0; pos < mpi_mean; ++pos) {
#else
#ifdef _OPENMP
		int dx = sc.seqLen();
#pragma omp parallel for private(categor)
		for (pos = 0; pos < dx; ++pos) {
#else
		for (pos = 0; pos < sc.seqLen(); ++pos) {
#endif
#endif
			MDOUBLE posPr = posLike[pos];
			for (categor = 0; categor < sp.categories(); ++categor) {
#ifdef _MPI
				cmargAlg.fillComputeMarginal(et,sc,sp,pos+mpi_spos,pij0[categor],cmarg[pos][categor],cup[pos][categor],cdown[pos][categor],posPr);
#else
				cmargAlg.fillComputeMarginal(et,sc,sp,pos,pij0[categor],cmarg[pos][categor],cup[pos][categor],cdown[pos][categor],posPr);
#endif
			}
		}
		timer_stop(2); // openMP parallel time 2
		para_time2 += timer_read(2);
		
#ifdef _OUTPUT	
		printf ("\nS-CUD para time: %.2fs\n", para_time2);
#endif

		semphyStep semphyStep1(
			  et,//tree& et,
			  sc,//const sequenceContainer1G& sc,
			  sp,//const stochasticProcess& sp,
			  pij0,//const computePijGam& pij0,
			  cup,//const suffStatGlobalGam& cup,
			  cdown,//const suffStatGlobalGam& cdown,
			  true,//const bool useApproxCounts,
			  posLike,//const vector<MDOUBLE>& cprobAtEachPos,
			  ratePosteriorProb, //CODE_RED
			  cmarg,//const suffStatGlobalGam& computeMarginal,
				NULL,//		const Vdouble *weights,
				0.001);//	const MDOUBLE toll);

		semphyStep1.doNothing(); // just so the compiler will not complain

		timer_stop(1);	// semphy step
		Semphy_time += timer_read(1);
		
		Semphy_para_time += timer_read(2);

#ifdef _OUTPUT	
		printf ("\nSemphy total time: %.2fs\n", Semphy_time);
#endif
	//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx finding the likelihood of the final trees, and output		

		
		timer_start(3);

		MDOUBLE resx =
				optimizeBranchLength1G_EM(et,sc,sp,NULL);

		timer_stop(3);	// BBL step
		BBL_time += timer_read(3);
		
#ifdef _OUTPUT	
		printf ("\nBBL total time: %.2fs\n", BBL_time);

		LOG(3,<<"likelihood after semphy step is: "<<resx<<endl);
		out<<"likelihood after semphy step is: "<<resx<<endl;  //ouyang
		
		LOG(3,<<"sep best tree under semphy (gam)"<<endl);
		LOGDO(3,et.output(myLog::LogFile()));
#endif		
		bool sameT = treeInterface::sameTreeTolopogy(bestTree,et);

#ifdef _EVAL
		// out<<"same Tree?"<<sameT<<endl;  //ouyang
#else
		if (sameT) break;
		if (resx < oldL + 0.05) break;
#endif
	}

#ifdef _OUTPUT	
	LOG(3,<<" ============== final tree and likelihoods ================== "<<endl);
	out<<" ============== final tree and likelihoods ================== "<<endl;
#endif
//	MDOUBLE resF = optimizeBranchLength1G_EM(et,sc,sp,NULL);

//	LOG(3,<<"likelihood after semphy: "<<resF<<endl);
//	out<<"likelihood after semphy: "<<resF<<endl;

#ifdef _OUTPUT	
	LOG(3,<<"Best trees under semphy (gam)"<<endl);
	out<<"Best trees under semphy (gam)"<<endl;
	LOGDO(3,et.output(myLog::LogFile()));
	et.output(out);
#endif
//	return resF;
}
