#include "semphyStep.h"
#include "computePijComponent.h"
#include "approxSemphyDistance.h"
#include "minSpanTree.h"
#include "rearangeTree.h"
#include "correctToCanonialForm.h"
#include "likelihoodComputation.h"
#include "brLenOpt.h"
#include "searchStatus.h"
#include "talRandom.h"
#include "computeUpAlg.h"
#include "computeDownAlg.h"
//#include "computeExactAlg.h"
//#include "computeMarginalAlg.h"
//#include "exactSemphyDistance.h"
//#include "jointTable.h"

#include "wtime.h"
extern MDOUBLE Semphy_para_time;

semphyStep::semphyStep(tree& et,
		  const sequenceContainer1G& sc,
		  const stochasticProcess& sp,
		  const computePijGam& pij0,
		  const suffStatGlobalGam& cup,
		  const suffStatGlobalGam& cdown,
		  const bool useApproxCounts,
		  const vector<MDOUBLE>& cprobAtEachPos,
		  const VVdouble & posteriorRateProbAtEachPos, //CODE_RED
		  const suffStatGlobalGam& computeMarginal,
				const Vdouble *weights,
				const MDOUBLE toll) : _et(et),_sc(sc),_sp(sp),_pij0(pij0),_cup(cup),_cdown(cdown),
				_useApproxCounts(useApproxCounts),_cprobAtEachPos(cprobAtEachPos),
				_posteriorRateProbAtEachPos(posteriorRateProbAtEachPos), //CODE_RED
				_computeMarginal(computeMarginal),_weights(weights),_toll(toll){
	computeSemphyStep();
}

void semphyStep::computeSemphyStep(){
	// for the starting tree decide which nodes are "real" and which are not.
	vector<char> isRealTaxa(_et.iNodes(),0);
	treeIterTopDownConst tIt(_et);
	for (tree::nodeP mynode = tIt.first(); mynode != tIt.end(); mynode = tIt.next()) {
		if (isLeaf(mynode)) {isRealTaxa[mynode->id()] = 1;}
	}

//	fillCpijUpDownExactMarginalProbAtEachPos();
	semphyDistance* semDis1 = computeQmatrix(_weights);// NULL = WEIGHTS...
	VVdouble penetlyTable = *(semDis1->getLikeDistanceTablePtr()); // copy
	MDOUBLE score = 0.0;
	rearrangeTree::pairSet inSet = minSpanTree::span_tree(penetlyTable,&score);

	maybePrintPenaltyTableAndStartingTree(penetlyTable); //(log output lvl 50)
	MDOUBLE qSTART = computeQstart(semDis1);
	maybePrintDistanceBetweenNodesTable(semDis1);//(log, 50)
	maybePrintSpanTreeListAndQAfterSpanTree(inSet,score);//(log, 50)

	rearrangeTree rearrangeTree1(&inSet,semDis1);
	rearrangeTree1.reconstructTree(_et);
	maybePrintTheTreeAfterRearrangeTree();//(log, 50)
	correctToCanonialForm ctcf(&_et,*semDis1->getDistanceTablePtr(),isRealTaxa);
	ctcf.correctTree();
	maybePrintTheTreeAfterCorrectToCanonialForm();//(log, 50)
	maybePrintQspanMinusQinit(score,qSTART);//(log, 50)
	if (semDis1!=NULL) delete semDis1;
}
/*
void semphyStep::fillCpijUpDownExactMarginalProbAtEachPos() {
	computeUpAlg::fillComputeUp(&_et,_pi,_computeUp1);
	_computeProbOfEachPos1->fillProbOfEachPosition(&_et,_pi,_computeUp1);
	computeDownAlg::fillDown(&_et,_pi,_computeUp1,_computeDown1);
	computeExactAlg::fillExact(&_et,_pi,_computeUp1,_computeDown1,_computeExact1);
	computeMarginalAlg::fillMarginal(&_et,_pi,_computeExact1,_computeProbOfEachPos1,_computeMarginal1);
}*/

void semphyStep::maybePrintPenaltyTableAndStartingTree(const VVdouble& penetlyTable) const{
	LOG(50,<< "start of semphy step"<<endl);
	LOGDO(50,_et.output(myLog::LogFile(),tree::ANCESTOR));
	LOGDO(50,_et.output(myLog::LogFile(),tree::PHYLIP));
	LOG(50, <<"START L = "<<likelihoodComputation::getTreeLikelihoodAllPosAlphTheSame(_et,_sc,_sp,NULL)<<endl);
	vector<tree::nodeP> allNodes;
	_et.getAllNodes(allNodes,_et.iRoot());
	for (int n1=0; n1 < allNodes.size(); ++n1) 
		LOG(50,<<allNodes[n1]->name()<<"\tid= "<<allNodes[n1]->id()<<endl);
	LOG(50,<<"the penetlyTable: "<<endl);
	for (int i=0; i < penetlyTable.size(); ++i) {
		for (int k=0; k < penetlyTable[i].size(); ++k) {
			LOG(50,<< penetlyTable[i][k]<<"\t");
		}
		LOG(50,<<endl);
	}
}

MDOUBLE semphyStep::computeQstart(const semphyDistance* inSemDist) const {
	MDOUBLE qSTART=0.0;
	vector<tree::nodeP> allNodes;
	_et.getAllNodes(allNodes,_et.iRoot());
	for (int n1=0; n1 < allNodes.size(); ++n1) {
		if (allNodes[n1]->father()!=NULL) qSTART += inSemDist->getLikeDistance(allNodes[n1]->id(),allNodes[n1]->father()->id());
	}
	LOG(50,<<" q start = "<<qSTART<<endl);
	return qSTART;
}
	
void semphyStep::maybePrintDistanceBetweenNodesTable(const semphyDistance* inSemDist) const  {
	if (50<=myLog::LogLevel()) {
		LOG(50,<< "printing the branch length table"<<endl);
		VVdouble tTable = *inSemDist->getDistanceTablePtr(); // copy
		for (int i=0; i < tTable.size(); ++i) {
			for (int k=0; k < tTable[i].size(); ++k) {
				LOG(50,<< tTable[i][k]<<"\t");
			}
			LOG(50,<<endl);
		}
	}
}

void semphyStep::maybePrintSpanTreeListAndQAfterSpanTree(
		const rearrangeTree::pairSet& inSet,const MDOUBLE score) const{
	if (50<=myLog::LogLevel()) {
		LOG(50,<<"printing span tree connection (id) set"<<endl);
		for (rearrangeTree::pairSet::const_iterator z= inSet.begin(); z!= inSet.end(); ++z) {
			LOG(50,<<(z->first)<<" is connected to "<< z->second<<endl);
		}
		LOG(50,<<endl);
		LOG(50,<<"the score of the span tree: "<<score<<endl);
	}
}	

void semphyStep::maybePrintTheTreeAfterRearrangeTree() const{
	LOG(50,<<"after rearrange tree"<<endl);
	LOGDO(50,_et.output(myLog::LogFile(),tree::ANCESTOR));
	LOGDO(50,_et.output(myLog::LogFile(),tree::PHYLIP));
}

void semphyStep::maybePrintTheTreeAfterCorrectToCanonialForm() const{
	LOG(50,<<"after correction to canonial form"<<endl);
	LOGDO(50,_et.output(myLog::LogFile(),tree::ANCESTOR));
	LOGDO(50,_et.output(myLog::LogFile(),tree::PHYLIP));
}

void semphyStep::maybePrintQspanMinusQinit(const MDOUBLE qpan, const MDOUBLE qstart) const{
	LOG(50,<<" q span - q start = "<<qpan-qstart<<endl);
}

semphyDistance* semphyStep::computeQmatrix(const Vdouble* weight) {
//	_searchStat1->createRandomWeightsGamma(_t1->seqLen());
	semphyDistance* semDis1 = NULL;
	if (_useApproxCounts) {
		//cerr<<" using APPROX! "<<endl;
		semDis1 =  new approxSemphyDistance(_et,
						_sc,
						_sp,
						_pij0,
						_cup,
						_cdown,
						_cprobAtEachPos,
						_posteriorRateProbAtEachPos, //CODE_RED
						_computeMarginal,
						_weights,
						_toll);
	}
	else {
		cerr<<" using EXACT!. To implement again... "<<endl;
		exit(4);
	  //semDis1 =  new exactSemphyDistance(_et,
		//			     *_pi,
		//			     *_computeUp1,
		//			     *_computeDown1,
		//			     *_computeProbOfEachPos1,
		//			     *_computeMarginal1,
		//			     weight,
		//				 _toll);
	}

	timer_clear(11);
	timer_start(11);	// Semphy parallel step;
	semDis1->computeDistances();
	timer_stop (11);
	Semphy_para_time += timer_read(11);
	
	return semDis1;
}

