/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//                INTEL CORPORATION PROPRIETARY INFORMATION                //
//   This software is supplied under the terms of a license agreement or   //
//  nondisclosure agreement with Intel Corporation and may not be copied   //
//   or disclosed except in accordance with the terms of that agreement.   //
//       Copyright (c) 2003 Intel Corporation. All Rights Reserved.        //
//                                                                         //
//  File:      pnlGaussianPotential.hpp                                    //
//                                                                         //
//  Purpose:   CGaussianPotential class definition                         //
//                                                                         //
//  Author(s):                                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
// GaussianPotential.hpp: interface for the GaussianFactor class.

///////////////////////////////////////////////////////////////////////////////

#ifndef __PNLGAUSSIANPOTENTIAL_HPP__
#define __PNLGAUSSIANPOTENTIAL_HPP__

#include "pnlPotential.hpp"
//#include "GaussianData.hpp"

PNL_BEGIN

class PNL_API CGaussianPotential : public CPotential  
{
public:
	static CGaussianPotential* Create( const intVector& domainIn,
                                       CModelDomain* pMD,
                                       int inMoment = -1,
                                       const floatVector& VecIn = floatVector(),
                                       const floatVector& MatIn = floatVector(),
                                       float normCoeff = 0.0f,
                                       const intVector& obsIndicesIn = intVector());
#ifdef PNL_OBSOLETE
	static CGaussianPotential* Create( const int *domain, int nNodes,
                                    CModelDomain* pMD,
                                    int inMoment = -1, float const* pVec = NULL,
                                    float const* pMat = NULL, 
                                    float normCoeff = 0.0f,
                                    const intVector& obsIndicesIn = intVector());
#endif
	static CGaussianPotential* Copy( const CGaussianPotential *pGauPot );
    static CGaussianPotential* CreateDeltaFunction( const intVector& domainIn,
                                       CModelDomain* pMD,
                                       const floatVector& meanIn ,
					                   int isInMoment = 1,
                                       const intVector& obsIndicesIn = intVector());
    static CGaussianPotential* CreateUnitFunctionDistribution( 
                                const intVector& domainIn, 
                                CModelDomain* pMD,
                                int isInCanonical = 1,
                                const intVector& obsIndicesIn = intVector()); 
    virtual CFactor* Clone() const;
    virtual CFactor* CloneWithSharedMatrices();
#ifdef PNL_OBSOLETE
	static CGaussianPotential* CreateDeltaFunction( const int *domain,
                    int nNodes, CModelDomain* pMD, const float *pMean,
                    int isInMoment = 1,
                    const intVector& obsIndicesIn = intVector());
	static CGaussianPotential* CreateUnitFunctionDistribution( 
                    const int *domain, int nNodes,CModelDomain* pMD,
                    int isInCanonical = 1,
                    const intVector& obsIndicesIn = intVector());
#endif    
    virtual float ProcessingStatisticalData( int nnumberOfEvidences);
	virtual void UpdateStatisticsEM(const CPotential *pMargPot, 
        const CEvidence *pEvidence = NULL);
    virtual void UpdateStatisticsML( const pConstEvidenceVector& evidencesIn );
#ifdef PNL_OBSOLETE
	virtual void UpdateStatisticsML(const CEvidence* const* pEvidencesIn, 
        int EvidenceNumber);
#endif
	virtual void Dump()const;
    void SetCoefficient( float coeff, int isForCanonical );
	float GetCoefficient( int forCanonical );
    virtual void GenerateSample( CEvidence* evidenceIO, int maximize = 0) const;
    virtual CPotential* ConvertStatisticToPot(int numOfSamples) const;
    virtual float GetLogLik( const CEvidence* pEv, const CPotential* pShrInfRes = NULL ) const;
    
    ~CGaussianPotential(){}
protected:
	CGaussianPotential( const int *domain, int nNodes, CModelDomain* pMD,
        const intVector& obsIndicesIn = intVector() );
	CGaussianPotential( const CGaussianPotential &pGauPot );
    CGaussianPotential( const CGaussianPotential* pGauPot );
private:
};

PNL_END

#endif // __PNLGAUSSIANPOTENTIAL_HPP__
