#include "pnlConfig.h"
#include "mpiWrap.h"

#ifdef _MPI
#include <stdio.h>
 
PNL_API int	_my_rank, _num_procs;
PNL_API int	_namelen;
PNL_API char	_processor_name[MPI_MAX_PROCESSOR_NAME];
PNL_API int	_version, _subversion;

PNL_API int	nTotalCases;

PNL_API void pnlMpiInit(int argc, char * argv[])
{
	MPI_CALL( MPI_Init(&argc, &argv) );
	MPI_CALL( MPI_Comm_rank(MPI_COMM_WORLD, &_my_rank) );
	MPI_CALL( MPI_Comm_size(MPI_COMM_WORLD, &_num_procs) );
	MPI_CALL( MPI_Get_processor_name(_processor_name,&_namelen) );

	MPI_CALL( MPI_Get_version(&_version, &_subversion) );

	MPI_REMARK( "[%d/%d] run at %s. MPI Ver %d.%d\n\n", 
	_my_rank, _num_procs, _processor_name, _version, _subversion );
}

PNL_API void pnlMpiFinalize()
{
	MPI_CALL( MPI_Finalize() );
}

//////////////////////////////////////////////////////////////
// the following code is only used in DLL

struct _tagRetCode
{
	int   code;
	char *msg;
};
       
_tagRetCode _RetCode[] = 
{
	{MPI_SUCCESS, "Successful return code"},
	{MPI_ERR_BUFFER, "Invalid buffer pointer"}, 
	{MPI_ERR_COUNT, "Invalid count argument"},
	{MPI_ERR_TYPE, "Invalid datatype argument"}, 
	{MPI_ERR_TAG, "Invalid tag argument"}, 
	{MPI_ERR_COMM, "Invalid communicator"}, 
	{MPI_ERR_RANK, "Invalid rank"}, 
	{MPI_ERR_ROOT, "Invalid root"}, 
	{MPI_ERR_GROUP, "Null group passed to function"}, 
	{MPI_ERR_OP, "Invalid operation"}, 
	{MPI_ERR_TOPOLOGY, "Invalid topology"}, 
	{MPI_ERR_DIMS, "Illegal dimension argument"}, 
	{MPI_ERR_ARG, "Invalid argument"}, 
	{MPI_ERR_UNKNOWN, "Unknown error"}, 
	{MPI_ERR_TRUNCATE, "message truncated on receive"}, 
	{MPI_ERR_OTHER, "Other error; use Error_string"}, 
	{MPI_ERR_INTERN, "internal error code"}, 
	{MPI_ERR_IN_STATUS, "Look in status for error value"}, 
	{MPI_ERR_PENDING, "Pending request"}, 
	{MPI_ERR_REQUEST, "illegal mpi_request handle"}, 
	{MPI_ERR_LASTCODE, "Last error code -- always at end"},
	{-1, "Emerging Error"}
};

char * _RetCode2String(int code)
{
	int i = 0;
	
	while(i <= MPI_ERR_LASTCODE)	
	{
		if (_RetCode[i].code == code)
			return _RetCode[i].msg;
	}
	return _RetCode[i].msg;
}
#endif
